import React from "react";

import FancyScrollbar from "../fancyscrollbar";
import Collapsible from "../collapsible";

import "./neighbors.css";
import { NodeType } from "../../../common/graph/nodetype";
import { Node } from "../../../common/graph/node";

interface NeighborsProps {
    neighbors: Node[];
    nodeClickedCallback?: (node: Node) => void;
}

/**
 * Shows a list of neighbors of a specific node separated by node type.
 * @param neighbors List of neighbors
 * @param nodeClickedCallback Optional callback to handle clicks on neighbors.
 * @param nodeColors Mapping of node types to colors.
 * @constructor
 */
function Neighbors({ neighbors, nodeClickedCallback }: NeighborsProps) {
    const classes = [...new Set<NodeType>(neighbors.map((node) => node.type))];
    // Sort classes to get a constant order of the node type tabs
    classes.sort((a, b) => {
        if (a.name < b.name) {
            return -1;
        }
        if (a.name > b.name) {
            return 1;
        }
        return 0;
    });

    const categories = new Map<number, Array<Node>>();

    for (const cls of classes) {
        categories.set(cls.id, []);
    }

    for (const neighbor of neighbors) {
        categories.get(neighbor.type.id).push(neighbor);
    }

    const handleNodeClick = (node: Node) => {
        if (nodeClickedCallback) {
            nodeClickedCallback(node);
        }
    };

    return (
        <div className={"neighbor-container"}>
            <FancyScrollbar>
                <Collapsible header={"Verwandte Inhalte"}>
                    {classes.map((cls) => (
                        <Collapsible
                            header={cls.name}
                            key={cls.id}
                            heightTransition={false}
                            color={cls.color}
                        >
                            <ul>
                                {categories.get(cls.id).map((node) => (
                                    <li
                                        className={"neighbor-content-link"}
                                        key={node.name}
                                        onClick={() => handleNodeClick(node)}
                                    >
                                        {node.name}
                                    </li>
                                ))}
                            </ul>
                        </Collapsible>
                    ))}
                </Collapsible>
            </FancyScrollbar>
        </div>
    );
}

export default Neighbors;
