import * as Config from "../config";
import * as THREE from "three";
import { ForceGraph3D } from "react-force-graph";

import { MODE, DRAG_THRESHOLD_3D } from "../config";
import background from "./background.jpg";
import alphaMask from "./alpha.png";

import { Line2, LineGeometry, LineMaterial } from "three-fatline";
import React from "react";
import PropTypes, { InferType } from "prop-types";
import SpriteText from "three-spritetext";
import { Object3D, Sprite } from "three";
import { Graph, Coordinate3D } from "../common/graph/graph";
import { Node } from "../common/graph/node";
import { Link } from "../common/graph/link";

export interface VisualGraphNode extends Node {
    __threeObj: THREE.Group;
}

export interface VisualGraphLink extends Link {
    __lineObj?: Line2;
}

// It is important to extend from React.PureComponent here to remove unnecessary re-renders!
// see https://www.robinwieruch.de/react-prevent-rerender-component/

/**
 * Renders contents of a Graph object as 3d-graph representation.
 */
export class GraphRenderer extends React.PureComponent<
    InferType<typeof GraphRenderer.propTypes>,
    InferType<typeof GraphRenderer.stateTypes>
> {
    props: InferType<typeof GraphRenderer.propTypes>;
    state: InferType<typeof GraphRenderer.stateTypes>;
    private forceGraph: React.RefObject<any>; // using typeof ForceGraph3d produces an error here...
    private highlightedNodes: Set<VisualGraphNode>;
    private highlightedLinks: Set<VisualGraphLink>;
    private hoverNode: VisualGraphNode;
    private node3dObjects: Map<number, THREE.Group>;

    static propTypes = {
        graph: PropTypes.instanceOf(Graph).isRequired,
        width: PropTypes.number.isRequired,
        height: PropTypes.number.isRequired,
        onNodeClicked: PropTypes.func,
        onNodeDeselected: PropTypes.func,
        isFullscreen: PropTypes.bool,
    };

    static stateTypes = {};

    constructor(props: InferType<typeof GraphRenderer.propTypes>) {
        super(props);

        this.reset();
        this.forceGraph = React.createRef();
    }

    reset() {
        this.highlightedNodes = new Set();
        this.highlightedLinks = new Set();
        this.node3dObjects = new Map<number, THREE.Group>();
        this.hoverNode = null;
    }

    componentDidMount() {
        this.addBackground();
    }

    addBackground() {
        const sphereGeometry = new THREE.SphereGeometry(20000, 32, 32);
        const loader = new THREE.TextureLoader();
        const planeMaterial = new THREE.MeshBasicMaterial({
            map: loader.load(background),
            side: THREE.DoubleSide,
        }); //THREE.BackSide
        const mesh = new THREE.Mesh(sphereGeometry, planeMaterial);
        mesh.position.set(0, 0, 0);
        //mesh.rotation.set(0.5 * Math.PI, 0, 0);

        this.forceGraph.current.scene().add(mesh);
    }

    drawNode(node: VisualGraphNode): Object3D {
        const group = new THREE.Group();

        const text = new SpriteText(node.name);
        text.color = "white";
        text.backgroundColor = "black";
        text.textHeight = 5;
        // text.padding = 2;
        text.borderRadius = 5;
        text.borderWidth = 3;
        text.borderColor = "black";
        text.translateY(12);
        text.material.opacity = 0.85;
        text.renderOrder = 999;
        group.add(text);

        // Draw node circle image
        const textureLoader = new THREE.TextureLoader();
        textureLoader.setCrossOrigin("anonymous");
        const imageAlpha = textureLoader.load(alphaMask);

        const material = new THREE.SpriteMaterial({
            //map: imageTexture,
            color: node.type.color,
            alphaMap: imageAlpha,
            transparent: true,
            alphaTest: 0.2,
            depthWrite: false,
            depthTest: false,
        });

        const sprite = new THREE.Sprite(material);
        //sprite.renderOrder = 999; // This may not be optimal. But it allows us to render the sprite on top of everything else.
        sprite.scale.set(...Config.NODE_SIZE);

        group.add(sprite);
        this.node3dObjects.set(node.id, group);
        return group;
    }

    drawLink(link: Link) {
        const colors = new Float32Array(
            [].concat(
                ...[link.target, link.source]
                    .map((node) => node.type.color)
                    .map((color) => color.replace(/[^\d,]/g, "").split(",")) // Extract rgb() color components
                    .map((rgb) => rgb.map((v: string) => parseInt(v) / 255))
            )
        );

        const geometry = new LineGeometry();
        geometry.setPositions([0, 0, 0, 1, 1, 1]);
        geometry.setColors(colors);

        const material = new LineMaterial({
            color: 0xffffff,
            linewidth: Config.LINK_WIDTH, // in world units with size attenuation, pixels otherwise
            vertexColors: true,

            resolution: new THREE.Vector2(
                window.screen.width,
                window.screen.height
            ), // Set the resolution to the maximum width and height of the screen.
            dashed: false,
            alphaToCoverage: true,
        });

        const line = new Line2(geometry, material);
        line.computeLineDistances();
        line.scale.set(1, 1, 1);

        return line;
    }

    onNodeHover(node: VisualGraphNode) {
        // no state change
        if (
            (!node && !this.highlightedNodes.size) ||
            (node && this.hoverNode === node)
        )
            return;

        const highlightedNodes = new Set<VisualGraphNode>();
        const highlightedLinks = new Set<VisualGraphLink>();

        if (node) {
            highlightedNodes.add(node);
            node.neighbors.forEach((neighbor) =>
                highlightedNodes.add(neighbor as VisualGraphNode)
            );
            node.links.forEach((link) =>
                highlightedLinks.add(link as VisualGraphLink)
            );
        }

        this.hoverNode = node || null;
        this.updateHighlight(highlightedNodes, highlightedLinks);
    }

    onLinkHover(link: VisualGraphLink) {
        const highlightedNodes = new Set<VisualGraphNode>();
        const highlightedLinks = new Set<VisualGraphLink>();

        if (link && link.__lineObj) {
            highlightedLinks.add(link);
            highlightedNodes.add(link.source as VisualGraphNode);
            highlightedNodes.add(link.target as VisualGraphNode);
        }

        this.updateHighlight(highlightedNodes, highlightedLinks);
    }

    updateHighlight(
        highlightedNodes: Set<VisualGraphNode>,
        highlightedLinks: Set<VisualGraphLink>
    ) {
        // Update Links
        this.highlightedLinks.forEach(
            (link) => (link.__lineObj.material.linewidth = Config.LINK_WIDTH)
        );
        this.highlightedLinks = highlightedLinks;
        this.highlightedLinks.forEach(
            (link) =>
                (link.__lineObj.material.linewidth = Config.HOVER_LINK_WIDTH)
        );

        // Update Nodes
        this.highlightedNodes.forEach((node) => {
            node.__threeObj.children[1].scale.set(...Config.NODE_SIZE);
        });
        this.highlightedNodes = highlightedNodes;
        this.highlightedNodes.forEach((node) => {
            node.__threeObj.children[1].scale.set(
                ...Config.HIGHLIGHTED_NODE_SIZE
            );
        });
    }

    deselectNode() {
        this.node3dObjects.forEach((nodeObj) => this.makeNodeVisible(nodeObj));

        for (const l of this.props.graph.links) {
            l.__lineObj.material.opacity = 1.0;
        }

        if (this.props.onNodeDeselected !== undefined) {
            this.props.onNodeDeselected();
        }
    }

    displayNodeSelection(node: VisualGraphNode) {
        // Set all nodes to a grayed out state
        this.node3dObjects.forEach((nodeObj) => {
            this.makeNodeInvisible(nodeObj);
        });

        // Reset neighbors
        for (const neighbor of node.neighbors) {
            const neighborObj = this.node3dObjects.get(neighbor.id);
            this.makeNodeVisible(neighborObj);
        }

        // Set all links to a grayed out state
        for (const l of this.props.graph.links) {
            l.__lineObj.material.opacity = 0.15;
        }

        // Reset for current connections
        for (const l of node.links) {
            const lObj = l as VisualGraphLink;
            lObj.__lineObj.material.opacity = 1.0;
        }

        // Reset for current node
        const nodeObj = this.node3dObjects.get(node.id);
        this.makeNodeVisible(nodeObj);
    }

    private makeNodeInvisible(nodeObjectGroup: THREE.Group) {
        const text = nodeObjectGroup.children[0] as SpriteText;
        text.material.opacity = 0.0;

        const circle = nodeObjectGroup.children[1] as Sprite;
        circle.material.opacity = 0.5;
    }

    private makeNodeVisible(nodeObjectGroup: THREE.Group) {
        const text = nodeObjectGroup.children[0] as SpriteText;
        text.material.opacity = 0.85;

        const circle = nodeObjectGroup.children[1] as Sprite;
        circle.material.opacity = 1.0;
    }

    onNodeClick(node: VisualGraphNode) {
        this.focusOnNode(node);
        if (MODE === "default" && this.props.onNodeClicked) {
            this.props.onNodeClicked(node);
        }
        this.displayNodeSelection(node);
    }

    onNodeDragEnd(node: VisualGraphNode, translate: Coordinate3D) {
        // NodeDrag is handled like NodeClick if distance is very short
        if (
            Math.hypot(translate.x, translate.y, translate.z) <
            DRAG_THRESHOLD_3D
        ) {
            this.onNodeClick(node);
        }
    }

    focusOnNode(node: VisualGraphNode) {
        const distance = 400; // Aim at node from outside it
        const speed = 0.5; // Camera travel speed through space
        const minTime = 1000; // Minimum transition time
        const distRatio = 1 + distance / Math.hypot(node.x, node.y, node.z);

        const currentPos = this.forceGraph.current.cameraPosition();
        const newPos = {
            x: node.x * distRatio,
            y: node.y * distRatio,
            z: node.z * distRatio,
        };
        const travelDist = Math.hypot(
            newPos.x - currentPos.x,
            newPos.y - currentPos.y,
            newPos.z - currentPos.z
        );

        this.forceGraph.current.cameraPosition(
            newPos,
            node, // lookAt ({ x, y, z })
            Math.max(travelDist / speed, minTime) // ms transition duration
        );
    }

    updateLinkPosition(line: Line2, start: Coordinate3D, end: Coordinate3D) {
        if (!(line instanceof Line2)) {
            return false;
        }
        // console.log("Updating links");
        const startR = 4;
        const endR = 4;
        const lineLen = Math.hypot(
            end.x - start.x,
            end.y - start.y,
            end.z - start.z
        );

        const positions = [startR / lineLen, 1 - endR / lineLen]
            .map((t) =>
                ["x", "y", "z"].map(
                    (dim) =>
                        start[dim as keyof typeof start] +
                        (end[dim as keyof typeof end] -
                            start[dim as keyof typeof start]) *
                            t
                )
            )
            .flat();
        line.geometry.setPositions(positions);

        return true;
    }

    render() {
        this.reset();
        return (
            <ForceGraph3D
                ref={this.forceGraph}
                width={this.props.width}
                height={this.props.height}
                graphData={this.props.graph}
                rendererConfig={{ antialias: true }}
                // nodeLabel={"hidden"}
                nodeThreeObject={(node: VisualGraphNode) => this.drawNode(node)}
                linkThreeObject={(link: Link) => this.drawLink(link)}
                onNodeClick={(node: VisualGraphNode) => this.onNodeClick(node)}
                onBackgroundClick={() => this.deselectNode()}
                //d3AlphaDecay={0.1}
                warmupTicks={150}
                cooldownTime={1000} // TODO: Do we want the simulation to unfreeze on node drag?
                enableNodeDrag={false}
                onNodeHover={(node: VisualGraphNode) => this.onNodeHover(node)}
                onLinkHover={(link: VisualGraphLink) => this.onLinkHover(link)}
                linkPositionUpdate={(
                    line: Line2,
                    coords: { start: Coordinate3D; end: Coordinate3D }
                ) => this.updateLinkPosition(line, coords.start, coords.end)}
                onNodeDragEnd={(
                    node: VisualGraphNode,
                    translate: Coordinate3D
                ) => this.onNodeDragEnd(node, translate)}
            />
        );
    }
}
